/*
    Astico2D: Another Simple Tool for Image Computing with OpenCV in 2D
    https://pageperso.lis-lab.fr/edouard.thiel/Astico2D/

    astico2d.hpp - main module 

    CC BY-SA Edouard.Thiel@univ-amu.fr - 30/07/2023 - v1.1
*/

#ifndef ASTICO2D_H
#define ASTICO2D_H

#include <iostream>
#include <iomanip>
#include <cstring>
#include <opencv2/opencv.hpp>


// Measure exec time; set to 0 if compilation problems
#define MEASURE_TIME 1

#if MEASURE_TIME
#  ifdef _WIN32
#    include <windows.h>
#  else 
#    include <sys/time.h>
#  endif

class AoMeasureTime {
  public:
    AoMeasureTime ();
    double get_time_dbl ();

  private:
#  ifdef _WIN32
    struct timeval _start_time;      // Initial date
    LARGE_INTEGER _start_count;      // Counter to compute time
    double _perf_freq;               // Frequency to compute time
    void _init_timeofday (void);
    void _get_systemtime (struct timeval *t);
    void _get_countertime (struct timeval *t);
#  endif
    void _gettimeofday (struct timeval *t);
};
#else // MEASURE_TIME
class AoMeasureTime {
  public:
    AoMeasureTime () {}
    double get_time_dbl () { return 0.; };
};
#endif // MEASURE_TIME


// This macro checks if an image has the expected type

#define CHECK_MAT_TYPE(mat, format_type) \
    if (mat.type() != int(format_type)) \
        throw std::runtime_error(std::string(__func__) +\
            ": format not handled '" + std::to_string(mat.type()) +\
            "' for matrix '" # mat "'")


// Colors management

void translate_to_vga_colors (cv::Mat &img_int, cv::Mat &img_col);
void reverse_img_colors (cv::Mat &img);
void reverse_pixel_colors (cv::Mat &img, int x, int y);
    

// Drawings

void draw_dotted_line (cv::Mat &img, const cv::Point &p1, 
    const cv::Point &p2, const cv::Scalar &color1, const cv::Scalar &color2, 
    const std::vector<bool> &pattern);
void draw_dotted_rectangle (cv::Mat &img, const cv::Point &p1, 
    const cv::Point &p2, const cv::Scalar &color1, const cv::Scalar &color2, 
    const std::vector<bool> &pattern);
void draw_dotted_cross (cv::Mat &img, const cv::Point &p1, 
    int l1, int l2, const cv::Scalar &color1, const cv::Scalar &color2, 
    const std::vector<bool> &pattern);


// Class AoText to draw text

class AoText {
  public:
    int fontFace = cv::FONT_HERSHEY_PLAIN;
    double fontScale = 1.0;
    int thickness = 1;
    cv::Scalar color = cv::Scalar::all(0);
    int lineType = cv::LINE_AA;  // LINE_4, LINE_8 or LINE_AA

    int baseline = 0;    // calculated
    int offset_h = 0;    // calculated

    void compute_size (std::string s, int &w, int &h);
    void draw_text (std::string s, int x, int y, cv::Mat &dest);
};


// Class AoPlace to manage the Map and the View

class AoPlace {
  public:
    int src_h = 100, src_w = 100;
    int zoom = 5, zoom_max = 100;
    int zoom_h = 500, zoom_w = 600;
    int zoom_x0 = 0, zoom_y0 = 0, zoom_x1 = 100, zoom_y1 = 100;
    int view_min = 300, view_max = 2000, view_inc = 50;
    int opt_h = 1000, opt_w = 1600;
    int map_w = 250, map_w_min = 150, map_w_max = 800, map_w_inc = 25;
    double map_f = 1.0;
    int footer_h = 20;
    std::string msg_left_footer, msg_right_footer;

    void init (int h, int w);
    void adjust_bounds ();
    void move (int dx, int dy);
    void resize_view (int dir_h, int dir_w);
    int  compute_optimal_zoom ();
    void resize_view_opt ();
    void get_cell_coords (int ev_x, int ev_y, int &cx, int &cy);
    void draw_cell_rect (cv::Mat &dest, int cx, int cy);
    void draw_cell_cross (cv::Mat &dest, int cx, int cy);
    void draw_location (cv::Mat &src, cv::Mat &dest);
    void draw_image_portion (cv::Mat &src, cv::Mat &dest);
    void draw_footer (cv::Mat &dest);
    void print_array_values (cv::Mat &src, int ex, int ey, int rx, int ry);
    void print_array_colors (cv::Mat &src, int ex, int ey, int rx, int ry);
    void compute_map_scale ();
    void change_map_width (int dir);
};


// Class Astico2D to manage UI

class Astico2D {
  public:
    int init_ok = true;
    std::string name_in1, name_out2, prog_name;
    cv::Mat img_src, img_res1, img_res2, img_int, img_col;
    int thresh_val = 127, thresh_max = 255;
    int zoom_slider = 5;  // caution, can be 0
    int click_x = 0, click_y = 0, click_dx = 0, click_dy = 0, click_n = 0,
        cell_x = -1, cell_y = -1;

    AoPlace place;

    enum RecomputeImages { R_NOTHING, R_VIEW, R_MAP, R_ALL };
    RecomputeImages recomp = R_ALL;

    char key_transform = 0;
    enum ModeTransform { M_NONE, M_LEVELS, M_COLORS };
    ModeTransform mode_transform = M_NONE;
    bool edit_mode = false;

    AoMeasureTime measure_time;

    Astico2D (int &argc, char **&argv);
    int run ();

    void print_usage ();
    virtual void print_keyboard_keys ();
    int analyse_args (int &argc, char **&argv);
    int read_input_image ();
    int save_result_image ();
    void create_internal_images ();
    void create_slider (const char* slider_name, int *current_value,
        int max_value);
    void modify_slider (const char* slider_name, int new_value);
    void decrease_slider (const char* slider_name, int min_value);
    void increase_slider (const char* slider_name, int max_value);

    void create_windows ();
    void rename_view_window (std::string name);
    void rename_map_window (std::string name);
    void set_left_footer (std::string msg);
    void set_right_footer (std::string msg);
    std::string compose_msg_map_coords ();

    void raise_recomp (RecomputeImages new_recomp);
    void handle_events ();
    virtual bool handle_keyboard (char key) { return false; };
    virtual void process_transformations () {};

  private:
    static void onZoomSlide (int pos, void *data);
    static void onDefaultTrackbarSlide (int pos, void *data);
    static void onMouseEventMap (int event, int x, int y, int flags, void *data);
    static void onMouseEventView (int event, int x, int y, int flags, void *data);
    static int onKeyPressEvent (int key, void *data);
};

#endif // ASTICO2D_H


